/* $XConsortium: sunCursor.c,v 5.19 94/04/17 20:29:35 gildea Exp $ */
/*

Copyright (c) 1988  Sun Microsystems, Inc.
Copyright (c) 1993  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/

/************************************************************
Copyright (c) 1994  Sun Microsystems, Inc.
Copyright (c) 1993  X Consortium

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documenta-
tion files (the "Software"), to deal in the Software with-
out restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to
whom the Software is furnished to do so, subject to the 
following conditions:

The above copyright notice and this permission notice shall
be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE, AND NONINFRINGEMENT.  IN NO EVENT SHALL SUN OR THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHER-
WISE, ARISING FROM, OUT OF, OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the names of Sun
Microsystems, Inc. and the X Consortium shall not be used in
advertising or otherwise to promote the sale, use, or other
dealings in this Software without prior written authoriza-
tion from Sun Microsystems, Inc. and/or the X Consortium, as
appropriate.
*************************************************************/

/*
 * modified for X11R6 Xvan server
 */

/*-
 * sunCursor.c --
 *	Functions for maintaining the Sun software cursor...
 *
 */

#define NEED_EVENTS
#include "sun.h"
#include "mfb.h"

#ifdef FBIOGCURMAX  /* has hardware cursor kernel support */

#define GetCursorPrivate(s) (&(GetScreenPrivate(s)->hardwareCursor))
#define SetupCursor(s)	    sunCursorPtr pCurPriv = GetCursorPrivate(s)

static void sunLoadCursor(ScreenPtr, CursorPtr, int, int);

static Bool
sunRealizeCursor (ScreenPtr pScreen, CursorPtr pCursor)
{
    SetupCursor(pScreen);
    int	    x, y;

    /* miRecolorCursor does this */
    if (pCurPriv->pCursor == pCursor)
    {
	miPointerPosition (&x, &y);
	sunLoadCursor (pScreen, pCursor, x, y);
    }
    return TRUE;
}

static Bool
sunUnrealizeCursor (ScreenPtr pScreen, CursorPtr pCursor)
{
    return TRUE;
}

static void
sunCursorRepad (ScreenPtr pScreen, CursorBitsPtr bits, 
		unsigned char *src_bits, unsigned int *dst_bits, 
		DDXPointPtr ptSrc, int w, int h)
{
  SetupCursor(pScreen);
  PixmapPtr src, dst;
  BoxRec box;
  RegionRec rgnDst;
  FBList src_fb, dst_fb;

  src_fb.add(bits->width, bits->height, 1, (unsigned int *) src_bits);
  dst_fb.add(w, h, 1, (unsigned int *) dst_bits);
  if (!(src = GetScratchPixmapHeader(pScreen, bits->width, bits->height,
				     /*bpp*/ 1, /*depth*/ 1,
				     PixmapBytePad(bits->width,1), 
				     &src_fb)))
    return;
  if (!(dst = GetScratchPixmapHeader(pScreen, w, h, /*bpp*/ 1, /*depth*/ 1,
				     PixmapBytePad(w,1), &dst_fb))) {
    FreeScratchPixmapHeader(src);
    return;
  }
  box.x1 = 0;
  box.y1 = 0;
  box.x2 = w;
  box.y2 = h;
  REGION_INIT(pScreen, &rgnDst, &box, 1);
  mfbDoBitblt((DrawablePtr) src, (DrawablePtr) dst, GXcopy, &rgnDst, ptSrc);
  REGION_UNINIT(pScreen, &rgnDst);
  FreeScratchPixmapHeader(src);
  FreeScratchPixmapHeader(dst);
}

static void
sunLoadCursor (ScreenPtr pScreen, CursorPtr pCursor, int x, int y)
{
  SetupCursor(pScreen);
  struct fbcursor fbcursor;
  int w, h;
  unsigned char r[2], g[2], b[2];
  DDXPointRec ptSrc;
/* The following was fixed so BUS signals don't occur when these
 * temp bitmaps are accessed with int pointer... PCJ 6-24-1993
  unsigned char source_temp[1024], mask_temp[1024];
 */
  unsigned int source_temp[256], mask_temp[256];
  fbFd *pfbfd;

  fbcursor.set = FB_CUR_SETALL;
  fbcursor.enable = 1;
  fbcursor.hot.x = pCursor->bits->xhot;
  fbcursor.hot.y = pCursor->bits->yhot;
  r[0] = pCursor->backRed >> 8;
  g[0] = pCursor->backGreen >> 8;
  b[0] = pCursor->backBlue >> 8;
  r[1] = pCursor->foreRed >> 8;
  g[1] = pCursor->foreGreen >> 8;
  b[1] = pCursor->foreBlue >> 8;
  fbcursor.cmap.index = 0;
  fbcursor.cmap.count = 2;
  fbcursor.cmap.red = r;
  fbcursor.cmap.green = g;
  fbcursor.cmap.blue = b;
  fbcursor.image = (char *) pCursor->bits->source;
  fbcursor.mask = (char *) pCursor->bits->mask;
  w = pCursor->bits->width;
  h = pCursor->bits->height;
  if (w > pCurPriv->width || h > pCurPriv->height) {
    ptSrc.x = 0;
    ptSrc.y = 0;
    if (w > pCurPriv->width)
      w = pCurPriv->width;
    if (h > pCurPriv->height)
      h = pCurPriv->height;
    sunCursorRepad (pScreen, pCursor->bits, pCursor->bits->source,
		    source_temp, &ptSrc, w, h);
    sunCursorRepad (pScreen, pCursor->bits, pCursor->bits->mask,
		    mask_temp, &ptSrc, w, h);
    fbcursor.image = (char *) source_temp;
    fbcursor.mask = (char *) mask_temp;
  }
  fbcursor.size.x = w;
  fbcursor.size.y = h;

  for (pfbfd = sunFbs[pScreen->myNum]; pfbfd; pfbfd = pfbfd->next) {
    fbcursor.pos.x = x - pfbfd->fbdesc->x;
    fbcursor.pos.y = y - pfbfd->fbdesc->y;
    /* if cursor is very far off this frame buffer, stash it in a known
     * location (just off bottom right), because cursors WAY off the screen
     * tend to reappear strangely...
     */
    if (fbcursor.pos.x < (0 - pCurPriv->width) ||
	fbcursor.pos.x >= (pfbfd->fbdesc->width + pCurPriv->width) ||
	fbcursor.pos.y < (0 - pCurPriv->height) ||
	fbcursor.pos.y >= (pfbfd->fbdesc->height + pCurPriv->height)) {
      fbcursor.pos.x = pfbfd->fbdesc->width + pCurPriv->width;
      fbcursor.pos.y = pfbfd->fbdesc->height + pCurPriv->height;
    }
    (void) ioctl (pfbfd->fd, FBIOSCURSOR, &fbcursor);
  }
}

static void
sunSetCursor (ScreenPtr pScreen, CursorPtr pCursor, int x, int y)
{
    SetupCursor(pScreen);

    if (pCursor)
    	sunLoadCursor (pScreen, pCursor, x, y);
    else
	sunDisableCursor (pScreen);
    pCurPriv->pCursor = pCursor;
}

static void
sunMoveCursor (ScreenPtr pScreen, int x, int y)
{
  SetupCursor(pScreen);
  struct fbcurpos pos;
  fbFd *pfbfd;

  for (pfbfd = sunFbs[pScreen->myNum]; pfbfd; pfbfd = pfbfd->next) {
    pos.x = x - pfbfd->fbdesc->x;
    pos.y = y - pfbfd->fbdesc->y;
    /* if cursor is very far off this frame buffer, stash it in a known
     * location (just off bottom right), because cursors WAY off the screen
     * tend to reappear strangely...
     */
    if (pos.x < (0 - pCurPriv->width) ||
	pos.x >= (pfbfd->fbdesc->width + pCurPriv->width) ||
	pos.y < (0 - pCurPriv->height) ||
	pos.y >= (pfbfd->fbdesc->height + pCurPriv->height)) {
      pos.x = pfbfd->fbdesc->width + pCurPriv->width;
      pos.y = pfbfd->fbdesc->height + pCurPriv->height;
    }
    (void) ioctl (pfbfd->fd, FBIOSCURPOS, &pos);
  }
}

miPointerSpriteFuncRec sunPointerSpriteFuncs = {
    (Bool (*)()) sunRealizeCursor,
    (Bool (*)()) sunUnrealizeCursor,
    (void (*)()) sunSetCursor,
    (void (*)()) sunMoveCursor,
};

static void
sunQueryBestSize (int class, unsigned short *pwidth, unsigned short *pheight,
		  ScreenPtr pScreen)
{
    SetupCursor (pScreen);

    switch (class)
    {
    case CursorShape:
	if (*pwidth > pCurPriv->width)
	    *pwidth = pCurPriv->width;
	if (*pheight > pCurPriv->height)
	    *pheight = pCurPriv->height;
	if (*pwidth > pScreen->width)
	    *pwidth = pScreen->width;
	if (*pheight > pScreen->height)
	    *pheight = pScreen->height;
	break;
    default:
	mfbQueryBestSize (class, pwidth, pheight, pScreen);
	break;
    }
}

extern miPointerScreenFuncRec	sunPointerScreenFuncs;

#endif

Bool sunCursorInitialize (ScreenPtr pScreen)
{
#ifdef FBIOGCURMAX
  SetupCursor (pScreen);
  struct fbcurpos maxsize;
  fbFd *pfbfd;

  pCurPriv->has_cursor = FALSE;
  for (pfbfd = sunFbs[pScreen->myNum]; pfbfd; pfbfd = pfbfd->next)
    if (ioctl (pfbfd->fd, FBIOGCURMAX, &maxsize) == -1)
      return FALSE;
  pCurPriv->width = maxsize.x;
  pCurPriv->height= maxsize.y;
  pScreen->QueryBestSize = sunQueryBestSize;
  miPointerInitialize (pScreen,
		       &sunPointerSpriteFuncs,
		       &sunPointerScreenFuncs,
		       FALSE);
  pCurPriv->has_cursor = TRUE;
  pCurPriv->pCursor = NULL;
  return TRUE;
#else
  return FALSE;
#endif
}

void sunDisableCursor (ScreenPtr pScreen)
{
#ifdef FBIOGCURMAX
  SetupCursor (pScreen);
  struct fbcursor fbcursor;
  fbFd *pfbfd;

  if (pCurPriv->has_cursor) {
    fbcursor.set = FB_CUR_SETCUR;
    fbcursor.enable = 0;
    for (pfbfd = sunFbs[pScreen->myNum]; pfbfd; pfbfd = pfbfd->next)
      (void) ioctl (pfbfd->fd, FBIOSCURSOR, &fbcursor);
    pCurPriv->pCursor = NULL;
  }
#endif
}
